@@ -23,21 +23,22 @@ from operator import methodcaller
 from os import fsdecode, remove, stat, stat_result, PathLike
 from pathlib import PurePosixPath
 from posixpath import join as joinpath
+from pprint import pformat
 from re import compile as re_compile, escape as re_escape, Pattern
 from shutil import copy, copyfileobj
 from tempfile import TemporaryDirectory
 from time import mktime
-from typing import Callable, Mapping, MutableMapping, Optional
+from typing import Any, Callable, Mapping, MutableMapping, Optional
 from types import MappingProxyType
 from uuid import uuid4
 from warnings import warn
 from weakref import WeakKeyDictionary, WeakValueDictionary
-from zipfile import ZipFile, ZIP_STORED
+from zipfile import ZipFile, ZIP_DEFLATED
 
 from .util.helper import items, wrap_open_in_bytes
 from .util.proxy import ElementAttribProxy, ElementProxy
-from .util.xml import el_add, el_del, el_find, el_iterfind, el_set, resolve_prefix
-from .util.undefined import undefined
+from .util.xml import el_add, el_del, el_iterfind, el_set
+from .util.undefined import undefined, UndefinedType
 
 try:
     from lxml.etree import fromstring, tostring, _Element as Element, _ElementTree as ElementTree # type: ignore
@@ -52,80 +53,6 @@ OPEN_MODES = frozenset(
 )
 
 
-def cache_meta_property(
-    initializer=undefined, 
-    instance_cache_attr=undefined, 
-    meta_attr=None, 
-    validator=None, 
-    _property_cls=type("property", (property,), {}), 
-):
-    if initializer is undefined:
-        return partial(
-            cache_meta_property, 
-            instance_cache_attr=instance_cache_attr, 
-            meta_attr=meta_attr, 
-            validator=validator, 
-        )
-    if callable(initializer):
-        if instance_cache_attr is undefined:
-            instance_cache_attr = initializer.__name__
-        def ensure_get(instance, /):
-            try:
-                return instance.__dict__[instance_cache_attr]
-            except LookupError:
-                el = instance.__dict__[instance_cache_attr] = initializer(instance)
-                return el
-        def pop_cache(self, /):
-            try:
-                el = self.__dict__.pop(instance_cache_attr)
-                warn("cache cleared")
-                return el
-            except LookupError:
-                pass
-    else:
-        if instance_cache_attr is undefined:
-            raise ValueError("if `initializer` is not a callable, `instance_cache_attr` must be specified")
-        def ensure_get(instance, /):
-            return getattr(instance, instance_cache_attr)
-        def pop_cache(self, /):
-            try:
-                el = getattr(self, instance_cache_attr)
-                delattr(self, instance_cache_attr)
-                warn("cache cleared")
-                return el
-            except AttributeError as e:
-                pass
-            except:
-                warn(f"can't clear cache: {instance_cache_attr!r}\n    |_ {type(e).__qualname__}: {e}")
-    def fget(self, /): 
-        if meta_attr is None:
-            return ensure_get(self).text
-        return ensure_get(self).get(meta_attr)
-    def fset(self, val, /):
-        if val is None:
-            pop_cache(self)
-        elif validator and not validator(self, val):
-            raise ValueError(f"invalid value: {val!r}")
-        elif meta_attr is None:
-            ensure_get(self).text = val
-        else:
-            ensure_get(self)[meta_attr] = val
-    def fdel(self, /):
-        el = pop_cache(self)
-        if el is not None:
-            try:
-                el._root.getparent().remove(el)
-            except:
-                pass
-    def set_validator(new_validator, /):
-        nonlocal validator
-        validator = new_validator
-        return prop
-    prop = _property_cls(fget, fset, fdel)
-    prop.set_validator = set_validator
-    return prop
-
-
 class FileInEpub:
     __slots__ = ("path", "internal", "open", "stat")
 
@@ -207,13 +134,28 @@ class Item(ElementAttribProxy):
         return joinpath(self.home, self.href)
 
     @property
-    def href(self):
+    def id(self, /):
+        return self._attrib["id"]
+
+    @property
+    def href(self, /):
         return self._attrib["href"]
 
     @href.setter
     def href(self, href_new, /):
         self.rename(href_new)
 
+    @property
+    def media_type(self, /):
+        return self._attrib["media-type"]
+
+    @media_type.setter
+    def media_type(self, value, /):
+        if not value:
+            self._attrib["media-type"] = guess_type(self._attrib["href"])[0] or "application/octet-stream"
+        else:
+            self._attrib["media-type"] = value
+
     @property
     def home(self, /):
         return elf._manifest._opf_dir
@@ -372,36 +314,83 @@ class Item(ElementAttribProxy):
 class Itemref(ElementAttribProxy):
     __const_keys__ = ("idref",)
 
+    @property
+    def idref(self, /):
+        return self._attrib["idref"]
+
+    @property
+    def linear(self, /):
+        return "no" if self._attrib.get("linear") == "no" else "yes"
+
+    @linear.setter
+    def linear(self, value, /):
+        self._attrib["linear"] = "no" if value == "no" else "yes"
+
 
 class Metadata(ElementProxy):
 
-    def add(self, name="meta", /, attrib=None, text=None, **_disregards):
+    def __repr__(self, /):
+        attrib = self._attrib or ""
+        return f"<{self.tag}>{attrib}\n{pformat(self.list())}"
+
+    def add(
+        self, 
+        name: str = "meta", 
+        /, 
+        attrib: Optional[Mapping] = None, 
+        text: Optional[str] = None, 
+        tail: Any = undefined, 
+        **_disregards, 
+    ):
         return super().add(name, attrib=attrib, text=text)
 
-    def dc(self, name, text_value=undefined, /, find_attrib=None, attrib=None, text=None, merge=False, delete=False, auto_add=True):
+    def dc(
+        self, 
+        name: str, 
+        text_value: UndefinedType | Optional[str] = undefined, 
+        /, 
+        find_attrib: Optional[Mapping] = None, 
+        attrib: Optional[Mapping] = None, 
+        text: Optional[str] = None, 
+        merge: bool = False, 
+        delete: bool = False, 
+        auto_add: bool = False, 
+    ):
         if text_value is not undefined:
             if find_attrib:
                 find_attrib = {**find_attrib, "": text_value}
             else:
                 find_attrib = {"": text_value}
-        return self.setfind("dc:%s" % name, find_attrib, attrib=attrib, text=text, merge=merge, delete=delete, auto_add=auto_add)
+        return self.setfind(
+            "dc:%s" % name, 
+            find_attrib=find_attrib, 
+            attrib=attrib, 
+            text=text, 
+            merge=merge, 
+            delete=delete, 
+            auto_add=auto_add, 
+        )
 
-    def meta(self, /, find_attrib=None, attrib=None, text=None, merge=False, delete=False, auto_add=True):
-        return self.setfind("{*}meta", find_attrib, attrib=attrib, text=text, merge=merge, delete=delete, auto_add=auto_add)
-
-    def name_meta(self, name, content=undefined, /, find_attrib=None, attrib=None, text=None, merge=False, delete=False, auto_add=True):
-        find_attrib = dict(find_attrib) if find_attrib else {}
-        find_attrib["name"] = name
-        if content is not undefined:
-            find_attrib["content"] = content
-        return self.meta(find_attrib, attrib=attrib, text=text, merge=merge, delete=delete, auto_add=auto_add)
-
-    def property_meta(self, property, text_value=undefined, /, find_attrib=None, attrib=None, text=None, merge=False, delete=False, auto_add=True):
-        find_attrib = dict(find_attrib) if find_attrib else {}
-        find_attrib["property"] = property
-        if text_value is not undefined:
-            find_attrib[""] = text_value
-        return self.meta(find_attrib, attrib=attrib, text=text, merge=merge, delete=delete, auto_add=auto_add)
+    def meta(
+        self, 
+        preds: str = "", 
+        /, 
+        find_attrib: Optional[Mapping] = None, 
+        attrib: Optional[Mapping] = None, 
+        text: Optional[str] = None, 
+        merge: bool = False, 
+        delete: bool = False, 
+        auto_add: bool = False, 
+    ):
+        return self.setfind(
+            "{*}meta%s" % preds, 
+            find_attrib=find_attrib, 
+            attrib=attrib, 
+            text=text, 
+            merge=merge, 
+            delete=delete, 
+            auto_add=auto_add, 
+        )
 
 
 class Manifest(dict):
@@ -1176,64 +1165,95 @@ class ePub(ElementProxy):
     def spine(self, /):
         return Spine(el_set(self._root, "{*}spine", "spine"), self.manifest)
 
-    @cache_meta_property
-    def identifier(self):
+    @property
+    def identifier(self, /):
         uid = self.get("unique-identifier")
         text = lambda: f"urn:uuid:{uuid4()}"
         if uid:
-            return self.metadata.dc("identifier", find_attrib={"id": uid}, text=text, merge=True)
+            return self.metadata.dc("identifier", find_attrib={"id": uid}, text=text, merge=True, auto_add=True).text
         else:
-            return self.metadata.dc("identifier", text=text, merge=True)
+            return self.metadata.dc("identifier", text=text, merge=True, auto_add=True).text
 
-    @cache_meta_property
-    def language(self):
-        return self.metadata.dc("language", text="en", merge=True)
+    @identifier.setter
+    def identifier(self, text, /):
+        uid = self.get("unique-identifier")
+        if uid:
+            self.metadata.dc("identifier", find_attrib={"id": uid}, text=text, auto_add=True)
+        else:
+            self.metadata.dc("identifier", text=text, auto_add=True)
 
-    @cache_meta_property
-    def title(self):
-        return self.metadata.dc("title", text="", merge=True)
+    @property
+    def language(self, /):
+        return self.metadata.dc("language", text="en", merge=True, auto_add=True).text
 
-    @cache_meta_property(meta_attr="content", validator=lambda self, cover_id: cover_id in self)
+    @language.setter
+    def language(self, text, /):
+        self.metadata.dc("language", text=text, auto_add=True)
+
+    @property
+    def title(self, /):
+        return self.metadata.dc("title", text="", merge=True, auto_add=True).text
+
+    @title.setter
+    def title(self, text, /):
+        self.metadata.dc("title", text=text, auto_add=True)
+
+    @property
     def cover(self, /) -> Optional[str]:
-        return self.metadata.name_meta("cover")
+        meta = self.metadata.meta('[@name="cover"]')
+        if meta is None:
+            return None
+        return meta.get("content")
+
+    @cover.setter
+    def cover(self, cover_id, /):
+        if cover_id not in self.manifest:
+            raise LookupError(f"no such item id: {cover_id!r}")
+        self.metadata.meta('[@name="cover"]', attrib={"content": cover_id}, auto_add=True)
 
     def pack(
         self, 
         /, 
         path=None, 
-        compression=ZIP_STORED, 
+        compression=ZIP_DEFLATED, 
         allowZip64=True, 
         compresslevel=None, 
     ):
         if not path and not self._path:
             raise OSError(errno.EINVAL, "please specify a path to save")
+        opf_dir, opf_name, opf_path = self._opf_dir, self._opf_name, self._opf_path
+        href_to_id = self.href_to_id
+        href_to_file = self.href_to_file
         def write_oebps():
-            bad_hrefs = []
-            for href, path in self.manifest._href_to_file.items():
-                if href == self._opf_name:
+            bad_ids = set()
+            good_ids = set()
+            for href, id in href_to_id.items():
+                if href == opf_name:
+                    bad_ids.add(id)
+                    warn(f"ignore a file, because its href conflicts with OPF: {href!r}")
+                    continue
+                file = href_to_file.get(href)
+                if file is None:
+                    bad_ids.add(id)
+                    warn(f"ignore a file, because it has never been actually created: {href!r}")
                     continue
                 try:
-                    fsrc = path.open("rb", buffering=0)
+                    fsrc = file.open("rb", buffering=0)
                 except (OSError, LookupError) as e:
+                    bad_ids.add(id)
                     warn(f"can't open {href!r}\n    |_ {type(e).__qualname__}: {e}")
-                    bad_hrefs.append(href)
-                    continue
                 else:
-                    with wfile.open(joinpath(self._opf_dir, href), "w") as fdst:
+                    with wfile.open(joinpath(opf_dir, href), "w") as fdst:
                         copyfileobj(fsrc, fdst)
+                    good_ids.add(id)
                 finally:
                     fsrc.close()
             root = self._root
-            if bad_hrefs:
+            if bad_ids:
                 root = deepcopy(root)
-                manifest = root[1]
-                for href in bad_hrefs:
-                    id = self.manifest._href_to_id[href]
-                    for item in manifest:
-                        if item.attrib["id"] == id:
-                            manifest.remove(item)
-                            break
-            wfile.writestr(self._opf_path, tostring(root, encoding="utf-8", xml_declaration=True))
+                manifest = root.find("{*}manifest")
+                manifest[:] = (item for item in manifest.iterfind("{*}item[@id]") if item.attrib["id"] in good_ids)
+            wfile.writestr(opf_path, tostring(root, encoding="utf-8", xml_declaration=True))
         zfile = self.__dict__.get("_zfile")
         if zfile is None:
             if path is None:
@@ -1251,20 +1271,19 @@ class ePub(ElementProxy):
         else:
             if path is None or path == self._path:
                 raise ValueError(f"not allowed to overwrite the original file: {self._path!r}")
+            exclude_files = {
+                joinpath(opf_dir, item.attrib["href"])
+                for item in fromstring(zfile.read(opf_path)).find("{*}manifest")
+            }
+            exclude_files.add(opf_path)
+            exclude_files.update(joinpath(opf_dir, href) for href in href_to_id)
             with ZipFile(path, "w", compression, allowZip64, compresslevel) as wfile:
-                exclude_files = {
-                    joinpath(self._opf_dir, item.attrib["href"])
-                    for item in fromstring(zfile.read(self._opf_path))[1]
-                } | {
-                    joinpath(self._opf_dir, href)
-                    for href in self.manifest._href_to_id
-                }
                 for name, info in zfile.NameToInfo.items():
-                    if info.is_dir() or name in exclude_files or name == self._opf_path:
+                    if info.is_dir() or name in exclude_files:
                         continue
                     with zfile.open(name) as fsrc, wfile.open(name, "w") as fdst:
                         copyfileobj(fsrc, fdst)
-                    write_oebps()
+                write_oebps()
 
 
 # TODO: 提供一个方法，对于 epub2 的电子书，可以把它的 guide、toc 等元素，等价转换并集中到 nav.xhtml 中
